# Test the bialternant formulas for symplectic and orthogonal Schur polynomials,
# with random rational numbers.

def random_rational_list(n):
   s = set()
   while len(s) < n:
      q = QQ.random_element()
      s.add(q)
   return list(s)

def prod_dif(x, j):
   return prod(x[j] - x[k] for k in range(j) + range(j + 1, len(x)))

def hom_via_geometric_progressions(x, m):
   n = len(x)
   return sum(x[j] ** (n + m - 1) / prod_dif(x, j) for j in range(n))


def chebyshev_V(n, a):
   return chebyshev_U(n, a) - chebyshev_U(n - 1, a) if n >= 1 else (a ** 0)

def chebyshev_W(n, a):
   return chebyshev_U(n, a) + chebyshev_U(n - 1, a) if n >= 1 else (a ** 0)

def sum_chebyshev_U(m, t):
   return sum([chebyshev_U(k, t) for k in range(m + 1)])


def homz(z, m):
   n = len(z)
   return sum(chebyshev_U(n + m - 1, z[j] / 2) / prod_dif(z, j) for j in range(n))

def homz_odd(z, m):
   n = len(z)
   return sum(sum_chebyshev_U(n + m - 1, z[j] / 2) / prod_dif(z, j) for j in range(n))

def spz_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k
         first_summand = homz(z, r) if r >= 0 else 0
         second_summand = homz(z, s) if s >= 0 else 0
         a = first_summand + second_summand
         M[j, k] = a if k > 0 else a / 2
   return det(M)

def oz_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz(z, r) if r >= 0 else 0
         second_summand = homz(z, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return det(M)

def oz_odd_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz_odd(z, r) if r >= 0 else 0
         second_summand = homz_odd(z, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return det(M)

def oz_odd_neg_via_jt1(z, la):
   zneg = [-z[j] for j in range(len(z))]
   sign_la = (-1) ** sum(la)
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz_odd(zneg, r) if r >= 0 else 0
         second_summand = homz_odd(zneg, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return sign_la * det(M)

def spz_odd_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k
         first_summand = homz_odd(z, r) if r >= 0 else 0
         second_summand = homz_odd(z, s) if s >= 0 else 0
         a = first_summand + second_summand
         M[j, k] = a if k > 0 else a / 2
   return det(M)

# bialternant formulas (``Schur-Chebyshov quotients'')

def partition_complemented(n, la):
   return la + [0] * (n - len(la))

def spz_via_bialternant(z, la): 
   n = len(z)
   la = partition_complemented(n, la)
   F = z[0].parent()
   Mnumer = matrix(F, n, n)
   Mdenom = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         pnumer = la[j] + n - j - 1
         pdenom = n - j - 1
         Mnumer[j, k] = chebyshev_U(pnumer, z[k] / 2)
         Mdenom[j, k] = z[k] ** pdenom
   return det(Mnumer) / det(Mdenom)

def oz_via_bialternant(z, la):
   n = len(z)
   la = partition_complemented(n, la)
   F = z[0].parent()
   Mnumer = matrix(F, n, n)
   Mdenom = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         pnumer = la[j] + n - j - 1
         pdenom = n - j - 1
         Mnumer[j, k] = 2 * chebyshev_T(pnumer, z[k] / 2) if pnumer > 0 else 1
         Mdenom[j, k] = z[k] ** pdenom
   return det(Mnumer) / det(Mdenom)

def oz_odd_via_bialternant(z, la):
   n = len(z)
   la = partition_complemented(n, la)
   F = z[0].parent()
   Mnumer = matrix(F, n, n)
   Mdenom = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         pnumer = la[j] + n - j - 1
         pdenom = n - j - 1
         Mnumer[j, k] = chebyshev_W(pnumer, z[k] / 2)
         Mdenom[j, k] = z[k] ** pdenom
   return det(Mnumer) / det(Mdenom)

def oz_odd_neg_via_bialternant(z, la):
   n = len(z)
   la = partition_complemented(n, la)
   F = z[0].parent()
   Mnumer = matrix(F, n, n)
   Mdenom = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         pnumer = la[j] + n - j - 1
         pdenom = n - j - 1
         Mnumer[j, k] = chebyshev_V(pnumer, z[k] / 2)
         Mdenom[j, k] = z[k] ** pdenom
   return det(Mnumer) / det(Mdenom)

def spz_odd_via_bialternant(z, la):
   n = len(z)
   la = partition_complemented(n, la)
   F = z[0].parent()
   Mnumer = matrix(F, n, n)
   Mdenom = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         pnumer = la[j] + n - j - 1
         pdenom = n - j - 1
         Mnumer[j, k] = sum_chebyshev_U(pnumer, z[k] / 2)
         Mdenom[j, k] = z[k] ** pdenom
   return det(Mnumer) / det(Mdenom)

# tests

def test_spz_bialternant_numer(n, la, verbos):
   if verbos >= 2:
      print('test_spz_bialternant_numer, n = %d, lambda = %s' % (n, str(la)))
   z = random_rational_list(n)
   res0 = spz_via_jt1(z, la)
   res1 = spz_via_bialternant(z, la)
   if verbos >= 3:
      print('spz via jt1         = ' + str(res0))
      print('spz via bialternant = ' + str(res1))
   return res0 == res1

def test_oz_bialternant_numer(n, la, verbos):
   if verbos >= 2:
      print('test_oz_bialternant_numer, n = %d, lambda = %s' % (n, str(la)))
   z = random_rational_list(n)
   res0 = oz_via_jt1(z, la)
   res1 = oz_via_bialternant(z, la)
   if verbos >= 3:
      print('oz via jt1         = ' + str(res0))
      print('oz via bialternant = ' + str(res1))
   return res0 == res1

def test_oz_odd_bialternant_numer(n, la, verbos):
   if verbos >= 2:
      print('test_oz_odd_bialternant_numer, n = %d, lambda = %s' % (n, str(la)))
   z = random_rational_list(n)
   res0 = oz_odd_via_jt1(z, la)
   res1 = oz_odd_via_bialternant(z, la)
   if verbos >= 3:
      print('oz odd via jt1         = ' + str(res0))
      print('oz odd via bialternant = ' + str(res1))
   return res0 == res1

def test_oz_odd_neg_bialternant_numer(n, la, verbos):
   if verbos >= 2:
      print('test_oz_odd_neg_bialternant_numer, n = %d, lambda = %s' % (n, str(la)))
   z = random_rational_list(n)
   res0 = oz_odd_neg_via_jt1(z, la)
   res1 = oz_odd_neg_via_bialternant(z, la)
   if verbos >= 3:
      print('oz odd neg via jt1         = ' + str(res0))
      print('oz odd neg via bialternant = ' + str(res1))
   return res0 == res1

def test_spz_odd_bialternant_numer(n, la, verbos):
   if verbos >= 2:
      print('test_spz_odd_bialternant_numer, n = %d, lambda = %s' % (n, str(la)))
   z = random_rational_list(n)
   res0 = spz_odd_via_jt1(z, la)
   res1 = spz_odd_via_bialternant(z, la)
   if verbos >= 3:
      print('spz odd via jt1         = ' + str(res0))
      print('spz odd via bialternant = ' + str(res1))
   return res0 == res1

# big tests

def big_test_spz_bialternant_numer(maxn, maxweight, nrep, verbos):
   if verbos >= 1:
      print('big_test_spz_bialternant_numer, maxn = %d, maxweight = %d, nrep = %d' % (maxn, maxweight, nrep))
   tasks = []
   for n in range(1, maxn + 1):
      for m in range(maxweight + 1):
         ps = sorted(list(Partitions(m, max_length = n)))
         tasks += [(n, la, r) for la in ps for r in range(nrep)]
   return all(test_spz_bialternant_numer(n, la, verbos) for (n, la, r) in tasks)

def big_test_oz_bialternant_numer(maxn, maxweight, nrep, verbos):
   if verbos >= 1:
      print('big_test_oz_bialternant_numer, maxn = %d, maxweight = %d, nrep = %d' % (maxn, maxweight, nrep))
   tasks = []
   for n in range(1, maxn + 1):
      for m in range(maxweight + 1):
         ps = sorted(list(Partitions(m, max_length = n)))
         tasks += [(n, la, r) for la in ps for r in range(nrep)]
   return all(test_oz_bialternant_numer(n, la, verbos) for (n, la, r) in tasks)

def big_test_oz_odd_bialternant_numer(maxn, maxweight, nrep, verbos):
   if verbos >= 1:
      print('big_test_oz_odd_bialternant_numer, maxn = %d, maxweight = %d, nrep = %d' % (maxn, maxweight, nrep))
   tasks = []
   for n in range(1, maxn + 1):
      for m in range(maxweight + 1):
         ps = sorted(list(Partitions(m, max_length = n)))
         tasks += [(n, la, r) for la in ps for r in range(nrep)]
   return all(test_oz_odd_bialternant_numer(n, la, verbos) for (n, la, r) in tasks)

def big_test_oz_odd_neg_bialternant_numer(maxn, maxweight, nrep, verbos):
   if verbos >= 1:
      print('big_test_oz_odd_neg_bialternant_numer, maxn = %d, maxweight = %d, nrep = %d' % (maxn, maxweight, nrep))
   tasks = []
   for n in range(1, maxn + 1):
      for m in range(maxweight + 1):
         ps = sorted(list(Partitions(m, max_length = n)))
         tasks += [(n, la, r) for la in ps for r in range(nrep)]
   return all(test_oz_odd_neg_bialternant_numer(n, la, verbos) for (n, la, r) in tasks)

def big_test_spz_odd_bialternant_numer(maxn, maxweight, nrep, verbos):
   if verbos >= 1:
      print('big_test_spz_odd_bialternant_numer, maxn = %d, maxweight = %d, nrep = %d' % (maxn, maxweight, nrep))
   tasks = []
   for n in range(1, maxn + 1):
      for m in range(maxweight + 1):
         ps = sorted(list(Partitions(m, max_length = n)))
         tasks += [(n, la, r) for la in ps for r in range(nrep)]
   return all(test_spz_odd_bialternant_numer(n, la, verbos) for (n, la, r) in tasks)

def all_small_tests_spz_oz_bialternant_formulas_numer():
   n = 2
   la = [3, 1]
   verbos = 3
   res = [0] * 5
   res[0] = test_spz_bialternant_numer(n, la, verbos)
   res[1] = test_oz_bialternant_numer(n, la, verbos)
   res[2] = test_oz_odd_bialternant_numer(n, la, verbos)
   res[3] = test_oz_odd_neg_bialternant_numer(n, la, verbos)
   res[4] = test_spz_odd_bialternant_numer(n, la, verbos)
   total_result = all(res)
   print('total result = ' + str(total_result))

import time

def all_big_tests_spz_oz_bialternant_formulas_numer():
   maxn = 5
   maxweight = 10
   nrep = 10
   verbos = 1
   t = time.time()
   res = [0] * 5
   res[0] = big_test_spz_bialternant_numer(maxn, maxweight, nrep, verbos)
   res[1] = big_test_oz_bialternant_numer(maxn, maxweight, nrep, verbos)
   res[2] = big_test_oz_odd_bialternant_numer(maxn, maxweight, nrep, verbos)
   res[3] = big_test_oz_odd_neg_bialternant_numer(maxn, maxweight, nrep, verbos)
   res[4] = big_test_spz_odd_bialternant_numer(maxn, maxweight, nrep, verbos)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

# time < 200 sec
all_big_tests_spz_oz_bialternant_formulas_numer()

